/**********************************************************************
 *
 * pipeout.c - External directory data processing. This module will
 * fork off external processes and send data to them on the standard
 * input.
 *
 * Data that can be sent is currently only the attributes of an entry
 * (in EDB format), but will be extended to a list of DNs.
 *
 *********************************************************************/

#include "defs.h"
#include "query.h"
#include "requests.h"
#include "pipeout.h"
#include "status.h"

#include <signal.h>
#include <errno.h>

#ifdef  BSD42
#include <sys/wait.h>
#endif

/**************************************************
 *
 * Local functions.
 * 
 */

static void ReadOutputAttributes();
static void GotOutputAttributes();

/* Fork a named process and pipe a character string to it. */
static BOOLEAN ExecAndPipeData();

/*****************************************************************************
 * XLUexternalDataProcess()
 *
 * Read the given attributes from the given entry and then pipe them
 * to the given executable.
 *
 ****************************************************************************/

void XLUexternalDataProcess(prim_widget,
			    entryname,
			    objid,
			    prim_id,
			    actionid,
			    trans)
     Widget prim_widget;
     String entryname;
     int objid;
     int prim_id;
     int actionid;
     XLUactionList trans;
{
  XLUprocessSpec pspec;

  pspec = (XLUprocessSpec) trans->param;
  if (pspec == (XLUprocessSpec) NULL)
    return;

  (void) ReadOutputAttributes(entryname,
			      pspec->attrlist,
			      prim_widget,
			      objid,
			      prim_id,
			      actionid,
			      trans);
  doPrimitiveSensitivity(objid);
} /* XLUexternalDataProcess */

/*****************************************************************************
 * ReadOutputAttributes()
 *
 * Retrieve a list of named attributes from an entry and return
 * as a dynamically allocated string.
 *
 ****************************************************************************/

static void ReadOutputAttributes(entryname,
				 attrlist,
				 prim_widget,
				 objid,
				 prim_id,
				 actionid,
				 trans)
     String entryname;
     stringList attrlist;
     Widget prim_widget;
     int objid;
     int prim_id;
     int actionid;
     XLUactionList trans;
{
  QyErrorCode err;
  int reqid;

  /* Request the attributes */
  err = QyReadEntry(entryname, &reqid, attrlist);

  /* Forget error reports for the moment! */
  switch (err)
    {
    case QERR_ok:
      break;

    default:
      return;
    }

  /* Record request invocation, giving a dummy object_id. */
  XLUrequestInvoked(reqid,
		    DS_READ,
		    GotOutputAttributes,
		    (DSReportCallback) NULL,
		    prim_widget,
		    objid,
		    prim_id,
		    actionid,
		    trans);
} /* ReadOutputAttributes */

/*****************************************************************************
 * GotOutputAttributes()
 *
 * A read request has completed. Get the attributes and pipe them
 * out to 
 *
 ****************************************************************************/

/* ARGSUSED */
static void GotOutputAttributes(reqid, objid, actionid, actlist)
     int reqid;
     int objid;
     int actionid;
     XLUactionList actlist;
{
  XLUprocessSpec pspec;
  QyReadResult result;
  PS outps;

  result = QyGetReadResult(reqid);
  XLUrequestComplete(reqid);

  doPrimitiveSensitivity(objid);
  pspec = (XLUprocessSpec) actlist->param;
  
  if (result->entry == NULLAVList)
    /* Read failed! */
    {
      XLUstatusMessage(objid, "Couldn't process data: directory read failed!");
    }
  else
    /* Read succeeded. */
    {
      QyAttrValList attr;
      stringCell value;

      outps = ps_alloc(str_open);
      if (outps == NULLPS)
	/* ### Error message */
	{
	  QyFreeReadResult(&result);
	  return;
	}

      if (str_setup(outps, NULLCP, 0, 0) == NOTOK)
	/* ### Error message */
	{
	  ps_free(outps);
	  QyFreeReadResult(&result);
	  return;
	}

      /* Print the attribute name and value pairs into a PStream. */
      for (attr = result->entry;
	   attr != (QyAttrValList) NULL;
	   attr = attr->next)
	{
	  for (value = attr->val_list;
	       value != NULLStrCell;
	       value = value->next)
	    {
	      ps_printf(outps,
			"%s= %s\n", attr->attr_name->string,
			value->string);
	    }
	}

      if (outps->ps_ptr != NULL)
	outps->ps_ptr = NULL;

      /* Should make and check a return on this. */
      (void) ExecAndPipeData(pspec->execname, outps->ps_base);
      ps_free(outps);
    }

  QyFreeReadResult(&result);
} /* GotOutputAttributes */

/*****************************************************************************
 * ExecAndPipeData()
 *
 * Fork a named process and pipe a character string to it.
 *
 ****************************************************************************/

static BOOLEAN ExecAndPipeData(execname, data)
     String execname;
     String data;
{
  int pd[2];
  char buffer[LINESIZE], execvector[LINESIZE];
  char *arg, *execfile;
  int pid, childpid;
  SFP pstat;
 
#ifndef BSD42
  int     status;
#else
  union wait status;
#endif /* BSD42 */
 
  extern int fork(), pipe(), close(), dup2(),
         _exit(), execl(), read(), write(), fflush();

  if (pipe(pd) == -1)
    {
      return FALSE;
    }

  pstat = signal(SIGPIPE, SIG_IGN);
  childpid = fork();
  switch (childpid)
    {
    case -1:
      
      (void) close (pd[1]);
      (void) close (pd[0]);
      (void) signal (SIGPIPE, pstat);
      
      return FALSE;
 
    case 0:
      
      (void) signal (SIGPIPE, pstat);
      
      if (dup2(pd[0], 0) == -1)
        _exit (-1);

      (void) close (pd[0]);
      (void) close (pd[1]);

      (void) sprintf(execvector, "%s", execname);
      execfile = TidyString(execvector);

      arg = index(execfile, ' ');
      if (arg != NULLCP)
	*arg++ = '\0';
      
      if (arg == NULLCP)
	(void) execlp(execvector, execvector, NULLCP);
      else
	(void) execlp(execfile, execfile, arg, NULLCP);
      
      while (read (0, buffer, sizeof buffer) > 0)
        continue;
      
      (void) printf ("ERROR: can't execute '%s'", execvector);
      (void) fflush (stdout);
 
      /* safety catch */
      _exit (-1);
 
      /* NOTREACHED */
 
    default:

      (void) close (pd[0]);
      if (write(pd[1], (char *) data, strlen(data)) != strlen(data))
        {
          (void) close (pd[1]);
          (void) signal (SIGPIPE, pstat);
 
          return FALSE;
        }
 
      (void) close (pd[1]);
 
      while ((pid = wait(&status)) != NOTOK && childpid != pid)
        continue;
      
      (void) signal (SIGPIPE, pstat);
    }

  return TRUE;
} /* ExecAndPipeData */
